from colorama import Fore, Back

fore = {
    "green": Fore.GREEN,
    "black": Fore.BLACK,
    "red": Fore.RED,
    "yellow": Fore.YELLOW,
    "blue": Fore.BLUE,
    "cyan": Fore.CYAN,
    "light_black": Fore.LIGHTBLACK_EX,
    "light_blue": Fore.LIGHTBLUE_EX,
    "light_cyan": Fore.LIGHTCYAN_EX,
    "light_green": Fore.LIGHTGREEN_EX,
    "light_magenta": Fore.LIGHTMAGENTA_EX,
    "light_red": Fore.LIGHTRED_EX,
    "light_white": Fore.LIGHTWHITE_EX,
    "light_yellow": Fore.LIGHTYELLOW_EX,
    "magenta": Fore.MAGENTA,
    "reset": Fore.RESET,
    "white": Fore.WHITE
}
"""This Is fore Dictionary That Contains Lots Of Colors That You Can Use For Text Color!"""

back = {
    "green": Back.GREEN,
    "black": Back.BLACK,
    "red": Back.RED,
    "yellow": Back.YELLOW,
    "blue": Back.BLUE,
    "cyan": Back.CYAN,
    "light_black": Back.LIGHTBLACK_EX,
    "light_blue": Back.LIGHTBLUE_EX,
    "light_cyan": Back.LIGHTCYAN_EX,
    "light_green": Back.LIGHTGREEN_EX,
    "light_magenta": Back.LIGHTMAGENTA_EX,
    "light_red": Back.LIGHTRED_EX,
    "light_white": Back.LIGHTWHITE_EX,
    "light_yellow": Back.LIGHTYELLOW_EX,
    "magenta": Back.MAGENTA,
    "reset": Back.RESET,
    "white": Back.WHITE
}
"""This Is back Dictionary That Contains Lots Of Colors That You Can Use For Background Of Text's Color!"""
